NAME crypto

; Copyright (C) 1997, 1998, 1999 Kevin Fu <fubob@mit.edu>
; Copyright (C) 1997, 1998, 1999 Ben Adida <ben@mit.edu>
; Copyright (C) 1997, 1998, 1999 Rodrigo Leroux
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2, or (at
; your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
; USA

;------------------
$INCLUDE	(../common/REG51FX.INC)
;$LIST
;$SYMLIST

PUBLIC makeKeyTable,crypt,puthex

EXTRN	CODE	(putchar)

crypto_code		SEGMENT CODE

RSEG crypto_code
; This makeKeyTable procedure assumes 128 bits=16 bytes of key
; Uses R0,R1,R3,R5
; The high byte of the address of the key
; table to initialize is in A


makeKeyTable:
	PUSH	PSW
	INC 	A
	MOV	R5, A	; R5 points to key table page
	MOV	P2, A	; Load key table page	

	; Initialize the key table
	MOV	R0, #0
InitKT:	 ;; fill 0-255
	MOV	A, R0
	MOVX	@R0, A
	DJNZ	R0, InitKT


	MOV	A,R5
	INC	A	; Load counter page
	MOV	P2,A

	; Initialize encryption counters i,j to zero
	MOV	A,#0
	MOV	R0,A
	MOVX	@R0,A	; i=0
	MOV	R0,#1
	MOVX	@R0,A	; j=0

	; Stir in the pasword
	; j=0
	; for i = 0 to 255 do
	;   j=(j+S_i+K_i) mod 256
	;   swap S_j,S_i


	; R0 is index1, R1 is index2
	MOV	R0, #0	; i
	MOV	R1, #0	; j
	
InitKT2:
	MOV	P2,R5	; Load key table page

	MOVX	A, @R0	; Get the ith byte, S_i, from the key table
	
	; A= S_i, R1= j
	ADD	A, R1	; j=S_i+j
	MOV	R1, A

	MOV	A, R5
	DEC	A	; Load password table page
	MOV	P2, A
	
	; save i
	MOV	A, R0
	MOV	R3, A
	
	ANL	A, #15	
	MOV	R0, A
	MOVX	A, @R0	; Get K_(i mod 16)

	; A = key_data_ptr[index1], R3= state[counter]+index2
	ADD	A, R1
	MOV	R1, A	; j=j+S_i+K_i
	MOV	A, R3
	MOV	R0, A

	; swap S_i,S_j

	MOV	P2, R5	; Load key table page
	
	MOVX	A, @R1	; Get S_j
	MOV	R3, A
	
	MOVX	A, @R0	; Get S_i

	; SWAP	S_i,S_j
	MOVX	@R1, A	; Set to S_i
	MOV	A, R3
	MOVX	@R0, A	; Set to S_j

	INC	R0	; i=i+1
	CJNE	R0, #0, InitKT2

doneKT:	POP	PSW
	RET


; Given: Value to encrypt is in A, address of password is in R3
; Returns: The encrypted byte in A.
; Note that this procedure is valid for encrypting or decrypting
; Stomps on R0,R1,R2,R3,R6
crypt:
	; i=i+1 mod 256
	; j=j+S_i mod 256
	; swap S_i,S_j
	; t=(S_i+S_j) mod 256
	; K=S_t
	; A=A XOR K

	MOV	R2,A	; Value to encrypt

	INC	R3	; Load counter page
	INC	R3
	MOV	P2,R3

	; Get the encryption counters i and j
	MOV	R0,#0
	MOVX	A,@R0
	INC	A	; i=i+1
	MOVX	@R0,A
	MOV	R0, A	; Save i for later

	DEC	R3	; Load key table page
	MOV	P2,R3
	
	MOVX	A, @R0	; Store S_i
	MOV	R6, A	
	
	INC	R3	; Load counter page
	MOV	P2,R3

	; get j
	MOV	R1,#1
	MOVX	A,@R1
	ADD	A, R6	; j=S_i+j
	MOVX	@R1,A
	MOV	R1, A	; Store j

	DEC	R3	; Load key table page
	MOV	P2,R3

	; SWAP S_i,S_j
	MOVX	A, @R1	
	MOVX	@R0, A	; Put S_j over the old S_i

	XCH	A, R6
	MOVX	@R1, A	; Put S_i over the old S_j
	
	ADD	A, R6	; t=S_i+S_j
	MOV	R0,A
	MOVX	A,@R0	; Get S_t
	
	; XOR random value S_t with stored byte to crypt
	XRL	A, R2
	
	RET

;; debugging procedures

; Prints 2 digits of hex number stored in A
; Does not use any regs (pushes, then pops).
puthex:
	PUSH	PSW
	PUSH	0x00	
	MOV	R0,A

	SWAP	A	; Just use high nibble
	ANL	A,#15

	CLR	C	
	SUBB	A,#10	
	JNC	alph		; if value was A-F, skip next step
	ADD	A,#'0'+10		; value was 0-9
	JMP	noalph
alph:	ADD	A,#'A'		; An alpha hex character
noalph:	CALL putchar 

	MOV	A,R0
	ANL	A,#15	; Just use low nibble

	CLR	C	
	SUBB	A,#10	
	JNC	alph2		; if value was A-F, skip next step
	ADD	A,#'0'+10		; value was 0-9
	JMP	noalph2
alph2:	ADD	A,#'A'		; An alpha hex character
noalph2:CALL putchar 
	MOV	A,R0
	POP	0x00
	POP	PSW
	RET







