; Copyright (C) 1997, 1998, 1999 Kevin Fu <fubob@mit.edu>
; Copyright (C) 1997, 1998, 1999 Ben Adida <ben@mit.edu>
; Copyright (C) 1997, 1998, 1999 Rodrigo Leroux
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2, or (at
; your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
; USA

NAME DIFFIEHELLMAN

; Assumes stack starts above bit-addressable regs

;------------------
$INCLUDE	(../common/REG51FX.INC)
$INCLUDE (..\common\macros.inc)

PUBLIC modexp, random, modp, getp, bigadd, bigsquare, multbyte

extern	code (puthex,putchar,putstrc)


dh_code		SEGMENT CODE

RSEG dh_code

GEN	EQU	3	; The public generator in Z_p*
FOO	EQU	10h	; crap space

; Given: R4 contains the high address of the exponent w/ 65 bytes & MSB=0
;		If A=0, then use GEN as base.  Otherwise use stuff in (A)
; Returns: Stores results of y=g^(R4) mod p in R3
; Munges: R0,R1,R2,A,R5
; exp takes a pointer to the high byte of external memory where
; the exponent, e, lives.  We then compute y=g^e mod p using
; modular exponentiation (approximately 512 multiplications)
; Eg, if R3=#90H, then the exponent lives in 0x9000-0x9039 (the first 64 bytes)
; and this program uses 0x9040-90C0 for scratch and the return value (the next 128 bytes)
; |.....(R4)=e......|.R/O copy of e.|
; |00             64 65          129
; |(R3)=g^(R4)mod p |..(original R3)|
;  FOO: ( original R3) |
; Algorithm sketch to compute y= a^b mod p:
; Set y=1
; For i=1 to length(b) do
;   y=y^2 mod p
;   If b_i=1 then y=y*a mod p
; B0 Indicator of which base mode (0=GEN, 1=(R3))
modexp:	SETB	0x20.1	; Indicator of if Y=1 (1=YES, 0=maybe no)

	MOV	R0,#129; Copy (R3) to the end of (R3) and R4 to end of R4
	MOV	R1,#64
mbaselp:MOV	P2,R3
	MOVX	A,@R1
	MOVX	@R0,A
	
	MOV	P2,FOO	; second copy
	MOVX	@R1,A

	MOV	P2,R4
	MOVX	A,@R1
	MOVX	@R0,A
	DEC	R1
	DEC	R0
	CJNE	R0,#64,mbaselp

	MOV	P2,R3	; init y=1
	MOV	R0,#0	;Address of the MSB of y
	CLR	A
	PUSH 	Acc

inity:	MOVX	@R0,A
	INC	R0
	CJNE	R0,#64,inity
	MOV	A,#1	; Set low byte of y to 1
	MOVX	@R0,A

	; Begin the exponentiation.  Repeat 512 times (once for each bit of (R3))
	; R0 and R1 are counters from 0 to 511
	MOV	R0,#0	; Indicates the current byte of the 64 bytes of e
	MOV	R1,#8	; Indicates the current bit of the current byte

	MOV	P2,R4
	MOVX	A,@R0	; Get the high byte of e
	CLR	C
modexpeachbit:
	; Check if y=1, if so, no need to square

	MOV	A,#'.'	
	CALL	putchar
	
	CJNE	R0,#64,modexpalmost
	CJNE	R1,#1,modexpalmost
	POP	Acc
	JMP	modexpdone
modexpalmost:
	JB	0x20.1,nosquareit

	MOV	A,#'s'
	CALL	putchar

	; Square the current y
	PUSH	0x01
	PUSH	0x00

	; Print operamds
	MOV	DPTR,#g2_msg
	CALL	putstrc
	MOV	P2,R3
	MOV	R0,#0
modyR2:MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, modyR2
	MOV	A,#'*'
	CALL	putchar
modx2:	MOVX	A,@R0
	CALL	puthex
	INC	R0
	CJNE	R0,#130,modx2


	XCH	A,R3	; Swap R3,R4
	XCH	A,R4
	XCH	A,R3
	CALL	bigsquare  ; R3<=R4^2
	XCH	A,R3	; Swap R3,R4 back
	XCH	A,R4
	XCH	A,R3
	
	; Move R4 to R3 and Restore R4
	MOV	R0,#64
	MOV	R1,#129
	MOV	P2,R4
restr4b:MOVX	A,@R0	; R3<-R4
	MOV	P2,R3
	MOVX	@R0,A
	MOV	P2,R4	; restore R4
	MOVX	A,@R1
	MOVX	@R0,A
	DEC	R0
	DEC	R1
	CJNE	R1,#64,restr4b

	; Print result
	MOV	DPTR,#square_msg
	CALL	putstrc
	MOV	P2,R3
	MOV	R0,#0
modyLL:	MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, modyLL

	POP	0x00
	POP	0x01


nosquareit:
	;; Decide whether to multiply by base
	DJNZ	R1,getnextebit	; Skip next step if we're on the same byte
	JMP	modexpnotdone
foo2:	JMP	modexpeachbit
modexpnotdone:
	INC	R0
	MOV	P2,R4
	POP	Acc
	MOVX	A,@R0
	PUSH	Acc
	MOV	R1,#8

	MOV	A,#'n'
	CALL	putchar

getnextebit:
	POP	Acc
	RLC	A	; The current bit is now in C
	PUSH	Acc

	JNC	foo2	; If the C=0, we're done
	
	CLR	0x20.1	; Y might not be 1 now.
	
	JNB	0x20.0,baseg
	
	MOV	A,#'y'
	CALL	putchar


	PUSH	0x00
	PUSH	0x01

	XCH	A,R3	; Swap R3,R4
	XCH	A,R4
	XCH	A,R3


	; Copy end of R5 over second of R4
	MOV	R0,#64
	MOV	R1,#129
damnlp:	MOV	P2,FOO
	MOVX	A,@R0
	MOV	P2,R4
	MOVX	@R1,A
	DEC	R0
	DEC	R1
	CJNE	R1,#64,damnlp



	; Print operamds
	MOV	DPTR,#g2_msg
	CALL	putstrc
	MOV	P2,R4
	MOV	R0,#0
modyLLLL2:MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, modyLLLL2
	MOV	A,#'*'
	CALL	putchar
modx:	MOVX	A,@R0
	CALL	puthex
	INC	R0
	CJNE	R0,#130,modx

	;; Multiply (R3)=(R4) *y
	CALL	bigmult


	XCH	A,R3	; Swap R3,R4 back
	XCH	A,R4
	XCH	A,R3

	; Move R4 to R3 and Restore R4
	MOV	R0,#64
	MOV	R1,#129
	MOV	P2,R4
restr4:	MOVX	A,@R0	; R3<-R4
	MOV	P2,R3
	MOVX	@R0,A
	MOV	P2,R4	; restore R4
	MOVX	A,@R1
	MOVX	@R0,A
	DEC	R0
	DEC	R1
	CJNE	R1,#64,restr4

	; Print result
	MOV	DPTR,#g_msg
	CALL	putstrc
	MOV	P2,R4
	MOV	R0,#0
modyLLLL:MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, modyLLLL

	POP	0x01
	POP	0x00

	JMP	modexpeachbit

baseg:	; Multiply y by g
	MOV	A,#'g'
	CALL	putchar

	MOV	R2,#GEN

	PUSH	0x01
	CALL	multbyte	; (R3)<-(R3)*G


	; Print result
;	MOV	DPTR,#g_msg
;	CALL	putstrc
;	MOV	P2,R3
;	MOV	R1,#0
;modyLLL:MOVX	A, @R1
;	CALL	puthex
;	INC	R1
;	CJNE	R1,#65, modyLLL

	POP	0x01

	JMP	modexpeachbit

modexpdone:
	; We're done.  Make sure we took mod p.
	CALL	modp

	RET


; Given: addresses of 65-byte numbers in R3 w/ MSB's=0
; Return: R3<-R4*Y' mod p
; Stomps on: R0,R1
; Assumes it can use bytes in R3 for scratch space
; R4: |...(Y)....|  (Y')    |..(Y R/O COPY).|
;      0       64 65     129 130         194
; R3: |..Y*Y'....|notouch   |
; Y <- Y*Y'
bigmult:
	; Make RO copy of Y and zero-out R3
	MOV	R0,#194
	MOV	R1,#64
bmloop:	MOV	P2,R4
	MOVX	A,@R1
	MOVX	@R0,A
	
	MOV	P2,R3	
	CLR	A
	MOVX	@R1,A	; Clear out R3
	DEC	R0
	DEC	R1
	CJNE	R0,#129,bmloop

	; Begin mult
	MOV	R0,#129
multop:	MOV	P2,R4
	MOVX	A,@R0

	MOV	R2,A

	MOV	A,R3	; Swap R3,R4
	XCH	A,R4
	MOV	R3,A

	CALL	multbyte	; (R3)<-(R3)*R2

	CJNE	R0,#129,mshiftit
	JMP	mnoshiftit
mshiftit:
	PUSH	0x00
	CLR	C
	MOV	A,R0
	SUBB	A,#65
	MOV	R0,A
	CALL	shift		; (R3)<-(R3)<<(64-(R0) bytes) mod p
				; (R3) <- (R3)<<(R0-129) == \

	POP	0x00

mnoshiftit:
	MOV	A,R3	; Swap R3,R4 back
	XCH	A,R4
	MOV	R3,A

	CALL	bigadd	; R3<- R3+R4 

	PUSH	0x00
	CALL	modp

	; Restore R4
	MOV	P2,R4
	MOV	R0,#194
	MOV	R1,#64
bmrest:	MOVX	A,@R0	
	MOVX	@R1,A
	DEC	R0
	DEC	R1
	CJNE	R0,#129,bmrest
	
	POP	0x00
	
	DEC	R0
	CJNE	R0,#64,multop

	RET

; Given: address of 65-byte numbers in R4 w/ MSB's=0 and 65-byte blank space in R3
; Return: R3<-R4^2 mod p
; Stomps on: R6,R0,R1
; Assumes it can use bytes in all of R3 for scratch space
;R3 |         (R3)         |
;    0                   64 65              129
;R4 |         (R4)         | (R4) RO copy      |
bigsquare:
	PUSH	PSW
	PUSH	Acc

	MOV	A,#0	; Init scratch space in R3 for accumulation
	MOV	R0,#65
	MOV	P2,R3
sqinit:	DEC	R0	
	MOVX	@R0,A
	CJNE	R0,#0,sqinit
	
	MOV	R0,#129	; Make a R/O copy of the contents pointed to by R4
	MOV	R1,#64
	MOV	P2,R4
sqcopy:	MOVX	A,@R1
	MOVX	@R0,A
	DEC	R0
	DEC	R1
	CJNE	R0,#64,sqcopy


	MOV	R0,#64
sqtop:	;MOV	A,#'.'
	;CALL	putchar
	MOV	P2,R4	; (R4) will always be restored to normal here
	MOVX	A,@R0

	MOV	R2,A

	MOV	A,R4	; Swap R3,R4
	XCH	A,R3
	MOV	R4,A

	CALL	multbyte	; (R3)<-(R3)*R2

	CJNE	R0,#64,shiftit
	JMP	noshiftit
shiftit:CALL	shift		; (R3)<-(R3)<<(64-(R0) bytes) mod p
	
noshiftit:
	MOV	A,R4	; Swap R3,R4 back
	XCH	A,R3
	MOV	R4,A

	CALL	bigadd	; R3<- R3+R4 mod p

	PUSH	0x00
	CALL	modp

	; Copy the original (R4) back to (R4)
	MOV	R0,#129	; Make a R/O copy of the contents pointed to by R4
	MOV	R1,#64
	MOV	P2,R4
sqcopy2:MOVX	A,@R0
	MOVX	@R1,A
	DEC	R0
	DEC	R1
	CJNE	R0,#64,sqcopy2
	POP	0x00
	
	DEC	R0
	CJNE	R0,#255,sqtop

	POP	Acc
	POP	PSW
	RET

; Given: R3 (65bytes w/ MSB=0) and R0 (byte shift ct b/w 1 and 63)
; Return: (R3)<-(R3)<<(64-(R0) bytes) mod p
; Stomps on: A,R1,R2
shift:	;MOV	A,#'<'
	;CALL	putchar
	MOV	A,#64
	CLR	C
	SUBB	A,R0	; Shift count in bytes
	JC	die	; debugging
	MOV	R2,A
	
shift2:	CALL	shiftabyte

	PUSH	0x00	; reduce it
	PUSH	0x02
	CALL	modp
	POP	0x02
	POP	0x00

	DJNZ	R2,shift2

	RET

; Given: R3 w/ MSB=0, mult by 256
; Munges: R1,A
shiftabyte:	; Move all the bytes in (R3) one to the left
	MOV	P2,R3		
	MOV	R1,#1
shiftabyteloop:
	MOVX	A,@R1
	DEC	R1
	MOVX	@R1,A
	INC	R1
	INC	R1
	CJNE	R1,#65,shiftabyteloop
	CLR	A
	DEC	R1
	MOVX	@R1,A
	RET

; Given: Address of 65-byte number in R3 (with MSB blank) and byte in R2 mod p
; Return: R3<-(R3)*R2 mod p, destroys R2
; Stomps: R1,R2,R5,A
multbyte:
	PUSH	0x00
	PUSH	0x04
	MOV	R0,#65	; Get LSB 
	MOV	P2,R3
	MOV	R5,#0	; carry byte

;	MOV	A,#'m'
;	CALL	putchar

mbloop:	DEC	R0	; trick to get around DJNZ


	MOVX	A,@R0
	MOV	B,R2
	MUL	AB
	ADD	A,R5
	MOVX	@R0,A
	MOV	A,B
	ADDC	A,#0	; There will never be any carry out
	JC	die	; debugging
	MOV	R5,A
	INC	R0

	DJNZ	R0,mbloop

	CALL	modp

	POP	0x04
	POP	0x00
	RET

die:	MOV	A,#' '
	CALL	putchar
	MOV	A,#'d'
	CALL	putchar
	MOV	A,#'i'
	CALL	putchar
	MOV	A,#'e'
	CALL	putchar
	JMP	$


; Given: Number of shifts in R0 (0-8), byte number in R1 (0-64)
; Return: In A, the R1th byte of the number P <<R0
;  		eg, If R0=0, R1=64, return the MSB of P
; P=0xa65554be82c14374159c45b66fbc8d520b50196397f8036969b410dc10d0276f5379b9915cf3652dace070f785f598c79b1f5097fb2abfaba7ca2a6ea79e0a2f
; Munges: R2,A
getp:
	PUSH	PSW
	;; Debugging, make sure R1 is between 0 and 64
	MOV	A,R1
	CLR	C
	SUBB	A,#65
	JNC	die

	;; debugging, make sure R0 is between 0 and 8
	MOV	A,R0
	CLR	C
	SUBB	A,#9
	JNC	die

	; Figure out how far to shift prime.
	CJNE	R0,#8,try7
	CALL	pr8
	POP	PSW
	RET
try7:	CJNE	R0,#7,try6
     	CALL	pr7
	POP	PSW
	RET
try6:	CJNE	R0,#6,try5
	CALL	pr6
	POP	PSW
	RET
try5:	CJNE	R0,#5,try4
	CALL	pr5
	POP	PSW
	RET
try4:	CJNE	R0,#4,try3
	CALL	pr4
	POP	PSW
	RET
try3:	CJNE	R0,#3,try2
	CALL	pr3
	POP	PSW
	RET
try2:	CJNE	R0,#2,try1
	CALL	pr2
	POP	PSW
	RET
try1:	CJNE	R0,#1,try0
	CALL	pr1
	POP	PSW
	RET
try0:	CALL	pr0
	POP	PSW
	RET

pr0:	MOV	DPTR,#prime0
	MOV	A,R1
	MOVC	A,@A+DPTR
	RET
prime0:	; P =	= P*16 >>4 P*256>>8  	P, a 65-byte prime
	DB	00h,0a6h,55h,54h,0beh,82h,0c1h,43h,74h,15h,9ch,45h,0b6h,6fh,0bch,8dh,52h,0bh
	DB	50h,19h,63h,97h,0f8h,03h,69h,69h,0b4h,10h,0dch,10h,0d0h,27h,6fh,53h,79h
	DB	0b9h,91h,5ch,0f3h,65h,2dh,0ach,0e0h,70h,0f7h,85h,0f5h,98h,0c7h,9bh,1fh,50h
	DB	97h,0fbh,2ah,0bfh,0abh,0a7h,0cah,2ah,6eh,0a7h,9eh,0ah,2fh,00h
        ; Dummy variable at the end


pr1:    MOV	DPTR,#prime1
	MOV	A,R1
	MOVC	A,@A+DPTR
	RET
prime1:	; P*2 = P*32 >>4  P shifted 1 bit
	DB	01h,4Ch,0AAh,0A9h,7Dh,05h,82h,86h,0E8h,2Bh,38h,8Bh,6Ch,0DFh,79h,1Ah,0A4h,16h,0A0h
	DB	32h,0C7h,2Fh,0F0h,06h,0D2h,0D3h,68h,21h,0B8h,21h,0A0h,4Eh,0DEh,0A6h,0F3h,73h,22h
	DB	0B9h,0E6h,0CAh,5Bh,59h,0C0h,0E1h,0EFh,0Bh,0EBh,31h,8Fh,36h,3Eh,0A1h,2Fh,0F6h,55h
	DB	7Fh,57h,4Fh,94h,54h,0DDh,4Fh,3Ch,14h,5Eh,00h


pr2:	MOV	DPTR,#prime2
	MOV	A,R1
	MOVC	A,@A+DPTR
	RET
prime2: ;P*4 = P*64 >>4
	DB	02h,99h,55h,52h,0FAh,0Bh,05h,0Dh,0D0h,56h,71h,16h,0D9h,0BEh,0F2h,35h,48h,2Dh,40h
	DB	65h,8Eh,5Fh,0E0h,0Dh,0A5h,0A6h,0D0h,43h,70h,43h,40h,9Dh,0BDh,4Dh,0E6h,0E6h,45h
	DB	73h,0CDh,94h,0B6h,0B3h,81h,0C3h,0DEh,17h,0D6h,63h,1Eh,6Ch,7Dh,42h,5Fh,0ECh,0AAh
	DB	0FEh,0AEh,9Fh,28h,0A9h,0BAh,9Eh,78h,28h,0BCh,00h


pr3:	MOV	DPTR,#prime3
	MOV	A,R1
	MOVC	A,@A+DPTR
	RET
prime3: ;P*8 = P*128 >>4
	DB	05h,32h,0AAh,0A5h,0F4h,16h,0Ah,1Bh,0A0h,0ACh,0E2h,2Dh,0B3h,7Dh,0E4h,6Ah,90h,5Ah
	DB	80h,0CBh,1Ch,0BFh,0C0h,1Bh,4Bh,4Dh,0A0h,86h,0E0h,86h,81h,3Bh,7Ah,9Bh,0CDh,0CCh
	DB	8Ah,0E7h,9Bh,29h,6Dh,67h,03h,87h,0BCh,2Fh,0ACh,0C6h,3Ch,0D8h,0FAh,84h,0BFh,0D9h
	DB	55h,0FDh,5Dh,3Eh,51h,53h,75h,3Ch,0F0h,51h,78h,00h

pr4:	CALL	pr0	; Get low nibble
	SWAP	A
	ANL	A,#0F0h
	MOV	R2,A
	INC	R1
	CALL	pr0
	SWAP	A
	ANL	A,#0Fh	; Get high nibble
	ORL	A,R2
	DEC	R1
	RET

pr5:	CALL	pr1	; Get low nibble
	SWAP	A
	ANL	A,#0F0h
	MOV	R2,A
	INC	R1
	CALL	pr1
	SWAP	A
	ANL	A,#0Fh	; Get high nibble
	ORL	A,R2
	DEC	R1
	RET
pr6:	CALL	pr2	; Get low nibble
	SWAP	A
	ANL	A,#0F0h
	MOV	R2,A
	INC	R1
	CALL	pr2
	SWAP	A
	ANL	A,#0Fh	; Get high nibble
	ORL	A,R2
	DEC	R1
	RET
pr7:	CALL	pr3	; Get low nibble
	SWAP	A
	ANL	A,#0F0h
	MOV	R2,A
	INC	R1
	CALL	pr3
	SWAP	A
	ANL	A,#0Fh	; Get high nibble
	ORL	A,R2
	DEC	R1
	RET
pr8:	INC	R1
	CALL	pr0
	DEC	R1
	RET

; Take a 65-byte number mod P
; The address of the number to reduce is in R3.
; Munges: R0,R1,R2
modp:
	MOV	P2,R3
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	; See if needs reducing 256*P
	CLR	C
	SUBB	A,#0A7h
	JC	red128	; no reduce

;	MOV	A,#'a'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#8	; Use P*256
	CALL	subp	; Subtract 256*P from number
	
red128:	; See if needs reducing 128*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#54h
	JC	red64	; no reduce

;	MOV	A,#'b'
;	CALL	putchar


	; Pass number as R2, pass prime shift in R0
	MOV	R0,#7	; Use P*128
	CALL	subp	; Subtract 128*P from number

red64:	; See if needs reducing 64*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#2Ah
	JC	red32	; no reduce


;	MOV	A,#'c'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#6	; Use P*64
	CALL	subp	; Subtract 64*P from number

red32:	; See if needs reducing 32*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#15h
	JC	red16	; no reduce


;	MOV	A,#'d'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#5	; Use P*32
	CALL	subp	; Subtract 32*P from number

red16:	; See if needs reducing 16*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#0Bh
	JC	red8	; no reduce


;	MOV	A,#'e'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#4	; Use P*16
	CALL	subp	; Subtract 16*P from number


red8:	; See if needs reducing 8*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#06h
	JC	red4	; no reduce

;	MOV	A,#'f'
;	CALL	putchar


	; Pass number as R2, pass prime shift in R0
	MOV	R0,#3	; Use P*8
	CALL	subp	; Subtract 8*P from number
red4:	; See if needs reducing 4*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#03h
	JC	red2	; no reduce


;	MOV	A,#'h'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#2	; Use P*4
	CALL	subp	; Subtract 4*P from number
red2:	; See if needs reducing 2*P
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte (MSB)

	CLR	C
	SUBB	A,#02h
	JC	red1	; no reduce


;	MOV	A,#'i'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#1	; Use P*2
	CALL	subp	; Subtract 2*P from number

red1:	; See if needs reducing P
	MOV	R1,#1
	MOVX	A,@R1	; Get 64th byte

	CLR	C
	SUBB	A,#0A7h
	JC	red0

;	MOV	A,#'j'
;	CALL	putchar


	; Pass number as R2, pass prime shift in R0
	MOV	R0,#0	; Use P
	CALL	subp	; Subtract P from number
red0:	; See if needs reducing P, take care of left over multiples
	MOV	P2,R3
	MOV	R1,#0
	MOVX	A,@R1	; Get 65th byte

	CLR	C
	SUBB	A,#01h
	JNC	redagain
	
	MOV	R1,#1
	MOVX	A,@R1	; Get 64th byte

	CLR	C
	SUBB	A,#0A7h
	JNC	redagain
	
	; Now, number to be reduced may at most start with A6.  Might need one more sub.
	; Check the MSB->LSB, making sure the prime byte>=input byte.  If match, enter 0
	MOV	R1,#1
	MOV	R0,#0
redlast:CALL	getp	; Get prime byte
	MOV	R2,A

	MOVX	A,@R1	; Get input byte
	
	CLR	C
	SUBB	A,R2
	JNC	redlastnext
	RET		; If R2>A, then return
redlastnext:
	CLR	C
	XCH	A,R2
	SUBB	A,R2
	JC	redlast2	; sub P

	INC	R1
	CJNE	R1,#65,redlast

	RET		; We're done
	
redagain:	; Keep sub'ing P
;	MOV	A,#'k'
;	CALL	putchar

	; Pass number as R3, pass prime shift in R0
	MOV	R0,#0	; Use P
	CALL	subp	; Subtract P from number
	JMP	red0

redlast2:	; Sub one P and return
;	MOV	A,#'l'
;	CALL	putchar

	; Pass number as R2, pass prime shift in R0
	MOV	R0,#0	; Use P
	CALL	subp	; Subtract P from number
	
	RET

; Given:  shift in R0 (in bits) and address of 65-byte number to subtract from in R3
; Compute (R3)<-(R3)-(P<<R0)
; Munges: R1, R2
subp:	PUSH	0x04
	MOV	R1,#64	; LSB->MSB 
	CLR	C
subnextbyte:
	CALL	getp
	MOV	R4,A	; Save prime byte
	MOV	P2,R3
	MOVX	A,@R1
	MOV	R2,A
	subbm	R2,R4	; R2=R2-R4
	MOV	A,R2
	MOVX	@R1,A
	DJNZ	R1,subnextbyte

	CALL	getp
	MOV	R4,A	; Save prime byte
	MOV	P2,R3
	MOVX	A,@R1
	MOV	R2,A
	subbm	R2,R4	; R2=R2-R4
	MOV	A,R2
	MOVX	@R1,A
	POP	0x04
	RET


; Given: 65-byte number at addr R3 (with MSB=0) and 65-byte num at addr R4 w/ MSB=0
; Return: (R3)<-(R3)+(R4) mod p, destroys R2
; Munges: R1, R2
; Works even if R3=R4
bigadd:	PUSH	0x00
	MOV	R1,#65	; LSB->MSB 
	CLR	C
addnextbyte:
	DEC	R1
	MOV	P2,R4
	MOVX	A,@R1
	MOV	R2,A	; Get next bytes to add

	MOV	P2,R3
	MOVX	A,@R1

	ADDC	A,R2	; Add 'em
	MOVX	@R1,A	; Store in R3
	
	INC	R1	; Trick to get around using CJNE which sets C flag!!!

	DJNZ	R1,addnextbyte

	;CALL	modp
	POP	0x00
	RET
; Given: Storage address in R3, store 65bytes of randomness with MSB=0
; Use the ADC's LSB to acquire 512 bits =64 bytes of randomness.
; Store results in external address pointed to by A
; Clobbers: R0,R1,R2,R4,R5
random:
	MOV	R1,#64	; Number of random bytes to collect (0..64)

nextrandombyte:
	MOV	R4,#80h	; The place to store random bits 
		; R4=1000:0000 (when 1 gets shifted out, the byte is ready.

nextrandombit:
	CLR	P1.7	; flip !convst				1
	SETB    P1.7    ; suppose it is connected to P1.7       1
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP		; Wait for ADC to convert
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP

	MOV	DPTR,#0
	MOVX	A,@DPTR

	RRC	A	; Get the LSB
	MOV	A,R4
	RRC	A	; Rotate that random bit into R4, rotate counter out of R4
	MOV	R4,A

	JNC	nextrandombit	; When the 1 in R4 gets shifted out, we've gathered 8 bits
	
	MOV	P2,R3	; Save random byte
	MOV	A, R4
	MOVX	@R1,A
	DJNZ	R1,nextrandombyte

	MOV	A,#0	; Set MSB=0
	MOVX	@R1,A


rag:	MOVX	@R1,A		; Blank out top 64 bytes
	INC	R1
	CJNE	R1,#64,rag

	CALL	modp	; reduce it

	RET


square_msg:
	DB	10,13,'square:',0
g_msg:
	DB	10,13,'gen:',0
g2_msg:
	DB	10,13,'operands:',0
