; Kryptophone
; Code that does voice encryption
; for phone line

; Copyright (C) 1997, 1998, 1999 Kevin Fu <fubob@mit.edu>
; Copyright (C) 1997, 1998, 1999 Ben Adida <ben@mit.edu>
; Copyright (C) 1997, 1998, 1999 Rodrigo Leroux
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2, or (at
; your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
; USA

NAME ENCRYPT

;------------------
$INCLUDE	(..\common\REG51FX.INC)
$LIST
$SYMLIST

EXTRN CODE (putchar,putstrc,getchar,gethex,puthex,crypt,makeKeyTable,wait_mode,modexp,random)

main_code		SEGMENT CODE
message_code		SEGMENT	CODE

CSEG AT 8000h
LJMP Start

RSEG main_code

B0	EQU	0x20.0	
B1	EQU	0x20.4
DHKE	EQU	0x20.2	; whether we have exchanged keys

SP_Init	EQU	29h	; system stack starts at 30
UART	EQU	4000h
EKEY	EQU	0A000h	; High byte only
DKEY	EQU	0A300h
RANDOMNESS	EQU	0A600h	;	512 bits of randomness
MODY	EQU	0A700h		;	g^RANDOMNESS mod p
MODX	EQU	0A800h		;	The random data from other party
CRAPSPACE	EQU	0AA00h
FOO	EQU	10h	; crap space

PREPARE_BYTE	EQU	0ffh
START_BYTE	EQU	0fh

; pins
DO_ENCRYPT_PIN	EQU	P1.2
DO_DECRYPT_PIN	EQU	P1.3
INITIATOR_PIN	EQU	P1.5

; Timer for 600 interrupts per second
TIMER_HIGH	EQU	59h
TIMER_LOW	EQU	00h

Start:
	MOV	SP,#SP_Init	; Set stack space
	CLR	B1
	CLR	DHKE	; (DHKE=0 -> Need to exchange keys)

	; Reset the UART
	; Write 0 to MCR
	MOV	A,#0
	MOV	DPTR,#4001H
	MOVX	@DPTR,A

	; Flush UART buffer
	MOV	DPTR, #UART
	MOVX	A, @DPTR

	; Read USR to reset
	MOV	DPTR,#4001H
	MOVX	A,@DPTR

	; Initialize the UART
	; Set the UART control regs
;	MOV	A,#48	; 8E1
;	MOV	A,#32	; 7E1
	MOV	A,#62	; 8N1
	MOV	DPTR,#4002H
	MOVX	@DPTR,A

	; Set the baud rate
	MOV	A,#0
	MOV	DPTR,#4003H
	MOVX	@DPTR,A

	; Set modem control regs
	MOV	A,#32
	MOV	DPTR,#4001H
	MOVX	@DPTR,A

	; Flush UART buffer
	MOV	DPTR, #UART
	MOVX	A, @DPTR

	MOV	DPTR,#init_message
	CALL	putstrc

	;; Get random number
	MOV	R3,#HIGH(RANDOMNESS)
	CALL	random

	MOV	DPTR,#random_msg
	CALL	putstrc

	; Print random number in HEX
	MOV	P2,#HIGH(RANDOMNESS)
	MOV	R0,#0
RLoopy:	MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, RLoopy

	; Compute	g^(RANDOMNESS) mod p
	MOV	DPTR,#mody_msg
	CALL	putstrc
	MOV	R3,#HIGH(MODY)
	MOV	R4,#HIGH(RANDOMNESS)
	MOV	FOO,#HIGH(CRAPSPACE)
	CLR	B0; Indicator of which base mode (0=GEN, 1=(R3))
	CALL	modexp

	; Print mody
	MOV	P2,#HIGH(MODY)
	MOV	R0,#0
modyL:	MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, modyL

encryptionInit:
 	CALL	wait_mode
	; Now we've done the exchange, we're going
	; to be sending 32 ff's, then one 0f, then
	; encrypting.
	; we need to do the same on the side of receiving.
	
	; Start timers
	MOV	TMOD, #22h
	MOV	TH0, #TIMER_HIGH
	SETB	TR0

	MOV	R7, #2fh; high nibble stands for state: 0=encrypt,1=rcv'd start,2-sending FF's
			; low nibble stands for remaining iterations to send FF's then start byte
encryptionloop:

	; wait for interrupt
	JNB	TF0, $	;wait
	CLR	TF0

;	MOV	A, #'-'
;	CALL	putchar

	; if we're still sending ff's
	MOV	A, R7
;	CALL	puthex
	ANL	A, #0fh
	JZ	send_for_real	
	DEC	R7
	JMP	send_prepare_byte

send_for_real:

;	MOV	A, #'*'
;	CALL	putchar

	; send i
	;MOV	P2, #HIGH(EKEY)+1H
	;MOV	R0, #0
	;MOVX	A, @R0

	;CALL	puthex

	;MOV	A, #'*'
	;CALL	putchar

	JNB	DHKE,exchange2

	CLR	P1.7	; flip !convst				1
	SETB    P1.7    ; suppose it is connected to P1.7       1

	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP
	NOP

	MOV	DPTR, #0h
	MOVX	A,@DPTR		; read A/D

;	MOV	R6, A
;	CALL	puthex
;	MOV	A, #'S'
;	CALL	putchar
;	MOV	A, R6

	MOV	R3,#HIGH(EKEY)	; address for i,j counters
	
	; save A away
	MOV	R4, A
	CALL	crypt
	
	; take out effects of encryption
	JB	DO_ENCRYPT_PIN, dont_restore_encrypt
	MOV	A, R4

dont_restore_encrypt:

	JMP	write_to_uart
	
send_prepare_byte:
	CJNE	A, #1, send_prepare_byte_for_real
	JMP	send_start_byte

send_prepare_byte_for_real:
	
	MOV	A, #PREPARE_BYTE
	JMP	write_to_uart
exchange2:	
	JMP	exchange
send_start_byte:
	MOV	DPTR,#start_msg
	CALL	putstrc

	MOV	A, #START_BYTE

write_to_uart:
	; write byte to UART
	MOV	DPTR, #UART
	MOVX	@DPTR, A

	;; HACK
;	MOV	R5, A
;	MOV	A, #'!'
;	CALL	putchar
;	MOV	A, R5

	;CALL	puthex
	;MOV	A, #'S'
	;CALL	putchar

	JNB	P1.4,$	; Wait until DR set

	;; HACK
;	MOV	A, #'+'
;	CALL	putchar

	; get a byte from the UART
	MOV	DPTR, #UART
	MOVX	A, @DPTR

;	CALL	puthex

; save that away for now
	MOV	R6, A

	;CALL	puthex
	;MOV	A, #'R'
	;CALL	putchar

	;; HACK
	;; JMP	do_decrypt_for_real

	; check if we're ready to decrypt
	MOV	A, R7
	SWAP	A
	ANL	A, #0fh
	JZ	do_decrypt_for_real

	; Here we check to see if we're getting PREPARE_BYTE, PREPARE_BYTE, START_BYTE
	CJNE	R6, #PREPARE_BYTE, check_start_byte

	MOV	A, #1
	JMP	done_handshake_iteration
	
check_start_byte:
	CJNE	R6, #START_BYTE, reset_wait_for_prepare
	MOV	A, #0

	JMP	done_handshake_iteration

reset_wait_for_prepare:	
	MOV	A, #2
	
done_handshake_iteration:
	SWAP	A
	XCH	A, R7
	ANL	A, #0fh
	ORL	A, R7
	MOV	R7, A
	JMP	encryptionloop

do_decrypt_for_real:

;	MOV	A, #'!'
;	CALL	putchar
	
	;MOV	A, #'('
	;CALL	putchar

	;MOV	P2, #HIGH(DKEY)+1H
	;MOV	R0, #0
	;MOVX	A, @R0
	
	;CALL	puthex

	;MOV	A, #')'
	;CALL	putchar

	; value is saved in R6
	MOV	A, R6
	MOV	R4, A
	
	MOV	R3,#HIGH(DKEY)	; address for i,j counters
	CALL	crypt
	
	; check if we shouldn't restore undecrypted version
	JB	DO_DECRYPT_PIN,dont_restore_decrypt
	MOV	A, R4

dont_restore_decrypt:

	; CALL	puthex

	MOV	DPTR, #0h
	MOVX	@DPTR, A

;	CALL	puthex
	
;	MOV	A,#'R'
;	CALL	putchar
	;MOV	A,#'-'
	;CALL	putchar

	JMP	encryptionloop

exchange:	;perform the Diffie-Hellman key exchange.
	; Given RANDOMNESS 	contains R
	;	MODY		contains g^R mod p
	; Obtain	MODX from other party
	; Compute	MODY=MODX^R mod p
	; If initiator, copy first half of MODY to EKEY, second half to DKEY
	; Otherwise copy other way around.

	MOV	DPTR,#starting_msg
	CALL	putstrc

	;Flush buffers
;repeatit:
;	MOV	DPTR,#UART
;	MOVX	A,@DPTR
;	JNB	TF0, $	;wait
;	CLR	TF0
;	JB	P1.4,repeatit

;	JNB	TF0, $	;wait a while
;	CLR	TF0
;	JNB	TF0, $	;wait
;	CLR	TF0
;	JNB	TF0, $	;wait
;	CLR	TF0

	; Send 65 bytes of information
	MOV	R0,#0
	MOV	A,#1
	MOV	DPTR,#MODY
	MOVX	@DPTR,A

;	JNB	INITIATOR_PIN,recvst
;	JMP	exlp
;recvst:; wait for interrupt
;	JNB	TF0, $	;wait
;	CLR	TF0
;	
;	JMP	reget
	
	JB	INITIATOR_PIN,exlp
	JNB	TF0, $	;wait
	CLR	TF0
	MOV	A,#'-'
	CALL	putchar

	MOV	P2,#HIGH(MODY)
	MOVX	A,@R0

	CALL	puthex
		
	; Send to other party
	MOV	DPTR,#UART
	MOVX	@DPTR,A

	INC	R0

exlp:	; wait for interrupt
	JNB	TF0, $	;wait
	CLR	TF0

	MOV	A,#'-'
	CALL	putchar

	MOV	P2,#HIGH(MODY)
	MOVX	A,@R0

	CALL	puthex
		
	; Send to other party
	MOV	DPTR,#UART
	MOVX	@DPTR,A

;	JB	INITIATOR_PIN,checkdone
;	JMP	notcheckdone
;checkdone:
;	CJNE	R0,#64,ibdone

;notcheckdone:
	MOV	A,#'+'
	CALL	putchar

reget:	JNB	P1.4,$	; Wait for DR

	MOV	DPTR,#UART
	MOVX	A,@DPTR
	
	JNB	B1,checkff
	JMP	nocheckff
checkff:
	DEC	A
	JZ	nocheckff2
	INC	A

	CLR	C
	SUBB	A,#START_BYTE
	JZ	chckstart
	ADD	A,#START_BYTE

	INC	A	; check ff
	JZ	reget2
	DEC	A


	; Mus be a receiver, inc R0
	SETB	B1	
;	INC	R0
	JMP	nocheckff

chckstart:
	MOV	DPTR,#UART
	MOV	A,#0h
	MOVX	@DPTR,A
	JMP	reget
reget2:	MOV	DPTR,#UART
	MOV	A,#0FFh
	MOVX	@DPTR,A
	JMP	reget
nocheckff2:	
	INC	A
	SETB	B1
nocheckff:
	PUSH	Acc

	MOV	A,#'('
	CALL	putchar
	MOV	A,R0
	CALL	puthex
	MOV	A,#')'
	CALL	putchar
	POP	Acc

	CALL	puthex

	MOV	P2,#HIGH(MODX)
	MOVX	@R0,A

	INC	R0
	CJNE	R0,#65,exlp



	; Zero out first byte
	MOV	P2,#HIGH(MODX)
	CLR	A
	MOV	R0,A
	MOVX	@R0,A

	; Compute	MODX=MODX^R mod p
	MOV	DPTR,#modexp_msg
	CALL	putstrc

	; Print modx
	MOV	P2,#HIGH(MODX)
	MOV	R0,#0
mexp1:	MOVX	A, @R0
	CALL	puthex
	INC	R0
	CJNE	R0,#65, mexp1


	SETB	B0; Indicator of which base mode (0=GEN, 1=(R3))
	MOV	R3,#HIGH(MODX)
	MOV	R4,#HIGH(RANDOMNESS)
	MOV	FOO,#HIGH(CRAPSPACE)
	CALL	modexp


	; Create EKEY and DKEY

	MOV	DPTR,#ktable_msg
	CALL	putstrc

	JB	INITIATOR_PIN,eskip1
	MOV	R0,#1
	JMP	eskip2
eskip1:	MOV	R0,#33
eskip2: MOV	R1,#0
get_Ekey:MOV	P2,#HIGH(MODX)
	MOVX	A,@R0
	MOV	R2,A
	INC	R0
	MOVX	A,@R0
	XRL	A,R2	; XOR two bytes of secret

	MOV	P2,#HIGH(EKEY)
	MOVX	@R1,A

	INC	R0
	INC	R1
	JNB	INITIATOR_PIN,eskip3
	JMP	eskip4
eskip3:	CJNE	R0,#33,get_Ekey
	JMP	eskipdone
eskip4:	CJNE	R0,#65,get_Ekey

eskipdone:	
	JB	INITIATOR_PIN,dskip1
	MOV	R0,#33
	JMP	dskip2
dskip1:	MOV	R0,#1
dskip2: MOV	R1,#0
get_Dkey:MOV	P2,#HIGH(MODX)
	MOVX	A,@R0
	MOV	R2,A
	INC	R0
	MOVX	A,@R0
	XRL	A,R2	; XOR two bytes of secret

	MOV	P2,#HIGH(DKEY)
	MOVX	@R1,A

	INC	R0
	INC	R1
	JB	INITIATOR_PIN,dskip3
	JMP	dskip4
dskip3:	CJNE	R0,#33,get_Dkey
	JMP	dskipdone
dskip4:	CJNE	R0,#65,get_Dkey

dskipdone:
	MOV	A,#'!'
	CALL	putchar

	;; Init key tables
	MOV	A,#HIGH(EKEY)
	CALL	makeKeyTable

	MOV	A,#HIGH(DKEY)
	CALL	makeKeyTable

; Print enc stuff
	MOV	DPTR,#enc_message
	CALL	putstrc

	MOV	DPTR, #EKEY
	MOV	R1,#16
printe:	MOVX	A, @DPTR
	CALL	puthex
	INC	DPTR
	DJNZ	R1,printe
	
	MOV	A,#10
	CALL	putchar
	MOV	A,#13
	CALL	putchar

	;; Print the key tables
	MOV	DPTR, #EKEY+100h

ELoopy:	MOVX	A, @DPTR
	CALL	puthex
	DJNZ	DPL, ELoopy
	MOV	A,#10
	CALL	putchar
	MOV	A,#13
	CALL	putchar

;; print dec stuff

	MOV	DPTR,#dec_message
	CALL	putstrc

	MOV	DPTR, #DKEY
	MOV	R1,#16
printd:	MOVX	A, @DPTR
	CALL	puthex
	INC	DPTR
	DJNZ	R1,printd
	
	MOV	A,#10
	CALL	putchar
	MOV	A,#13
	CALL	putchar

	MOV	DPTR, #DKEY+100h

DLoopy:	MOVX	A, @DPTR
	CALL	puthex
	DJNZ	DPL, DLoopy
	MOV	A,#10
	CALL	putchar
	MOV	A,#13
	CALL	putchar


	; Notify completion to other machine
	MOV	DPTR,#UART
	MOV	A,#0FFh
	MOVX	@DPTR,A

	JNB	P1.4,$;		Wait for DR

	MOVX	A,@DPTR
	CLR	C
	SUBB	A,#255
	JC	big_fat_error

	MOV	DPTR,#exchange_msg
	CALL	putstrc

	SETB	DHKE	; Signal done with DHKE
;	JMP	encryptionloop
	; Reset the UART
	; Write 0 to MCR
	MOV	A,#0
	MOV	DPTR,#4001H
	MOVX	@DPTR,A

	; Flush UART buffer
	MOV	DPTR, #UART
	MOVX	A, @DPTR

	; Read USR to reset
	MOV	DPTR,#4001H
	MOVX	A,@DPTR

	; Initialize the UART
	; Set the UART control regs
;	MOV	A,#48	; 8E1
;	MOV	A,#32	; 7E1
	MOV	A,#62	; 8N1
	MOV	DPTR,#4002H
	MOVX	@DPTR,A

	; Set the baud rate
	MOV	A,#0
	MOV	DPTR,#4003H
	MOVX	@DPTR,A

	; Set modem control regs
	MOV	A,#32
	MOV	DPTR,#4001H
	MOVX	@DPTR,A

	; Flush UART buffer
	MOV	DPTR, #UART
	MOVX	A, @DPTR
	JMP	encryptionInit

big_fat_error:
	MOV	DPTR,#big_fat_error_msg
	CALL	putstrc
	JMP	$

RSEG message_code

init_message:
	DB	10,13,'Kryptophone ready, UART initialized.',10,13,0
mode_message:
	DB	10,13,'(0) Sender or (1) Receiver: ',10,13,0
Ekey_message:
	DB	10,13,'Enter 128-bit encryption key (32 HEX chars in CAPS): ',0
Dkey_message:
	DB	10,13,'Enter 128-bit decryption key (32 HEX chars in CAPS): ',0
enc_message:
	DB	10,13,'Encrypting with: ',0
dec_message:
	DB	10,13,'Decrypting with: ',0
receive_message:
	DB	10,13,'RECEIVER started',10,13,0
send_message:
	DB	10,13,'SENDER started',10,13,0
big_fat_error_msg:
	DB	10,13,'BIG FAT ERROR in handshaking. Stopping Dead',10,13,0	
random_msg:
	DB	10,13,'Random exponent:',0
mody_msg:
	DB	10,13,'g^r mod p:',0
modexp_msg:
	DB	10,13,'Secure shared secret:',0
exchange_msg:
	DB	10,13,'Diffie-Hellman key exchange successful!',10,13,0
starting_msg:
	DB	10,13,'Starting Diffie-Hellman Exchange',10,13,0
start_msg:
	DB	10,13,'Sending start byte',10,13,0
ktable_msg:
	DB	10,13,'Creating Key Tables',10,13,0
END
