; Copyright (C) 1997, 1998, 1999 Kevin Fu <fubob@mit.edu>
; Copyright (C) 1997, 1998, 1999 Ben Adida <ben@mit.edu>
; Copyright (C) 1997, 1998, 1999 Rodrigo Leroux
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2, or (at
; your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
; USA

NAME HANDSHAKE


;------------------
$INCLUDE	(..\common\REG51FX.INC)
$LIST
$SYMLIST

PUBLIC	wait_mode

EXTRN CODE	(putchar, putstrc)

handshake_code	SEGMENT CODE

; This wait mode is the receive state
; this checks the initiator pin at each loop,
; also reading from the UART each time to
; see if a Hello message comes around

RSEG	handshake_code

DATA_READY	EQU	P1.4
INITIATOR_PIN	EQU	P1.5
UART		EQU	4000h

wait_mode:
	MOV	R4, #0
	MOV	R5, #32

	MOV	TMOD, #21h
	MOV	TH0, #80h
	MOV	TL0, #00h

	SETB	TR0

	MOV	DPTR, #receiver_mode_str
	CALL	putstrc

wait_loop:
	; wait for interrupt
	JNB	TF0, $
	CLR	TF0
	MOV	TH0, #80h
	MOV	TL0, #00h

	; If we already started recognizing an init
	; signal, then don't check the initiate pin
	CJNE	R4, #0, no_check_pin

	; Check the initate pin
	JB	INITIATOR_PIN, initiate_mode

no_check_pin:
	; Check if there is data waiting
	JNB	DATA_READY, no_data_ready
	
	; Get value from UART
	MOV	DPTR, #UART
	MOVX	A, @DPTR
	
	; Save it away
	MOV	R6, A
	
	; Check the value of the byte read
	MOV	A, R4
	MOV	DPTR, #hello_msg
	MOVC	A, @A+DPTR
	
	; If the value of the byte is good, increase
	; the counter of where we are in the hello msg
	CJNE	A, 0x06, byte_no_good
	INC	R4
	MOV	A, R4
	ADD	A, #32
	CALL	putchar
	MOV	A,#'.'
	CALL	putchar

	; If we have 8 good values for the hello
	; message, we have received a full good HELLO.
	; otherwise, we keep waiting for the hello
	CJNE	R4, #8, wait_loop
	JMP	send_response_message

no_data_ready:
	DJNZ	R5, wait_loop
	MOV	A,#'w'
	CALL	putchar
	
byte_no_good:
	; reset the counters
	MOV	R4, #0
	MOV	R5, #32
	
	MOV	A,#'q'
	CALL	putchar

	JMP	wait_loop

; This mode sends out the bytes of the initiator
; It does so in repetition, while waiting for the response
; message
initiate_mode:
	MOV	R3, #0
	MOV	R4, #0
	MOV	R5, #0

	MOV	DPTR, #initiator_mode_str
	CALL	putstrc

initiate_loop:
	; We wait for interrupt
	JNB	TF0, $
	CLR	TF0
	MOV	TH0, #80h
	MOV	TL0, #00h
	
	; Get the byte to send
	MOV	DPTR, #hello_msg
	MOV	A, R3
	MOVC	A, @A+DPTR
	
	MOV	DPTR, #UART
	MOVX	@DPTR, A

	INC	R3
	MOV	A, R3
	ANL	A, #7
	MOV	R3, A

	; Check if something has come in
	JNB	DATA_READY, no_receive_byte_ready

	; Get the byte we're supposed to receive
	MOV	DPTR, #respond_msg
	MOV	A, R4
	MOVC	A, @A+DPTR
	
	; Save byte 
	MOV	R2, A

	MOV	DPTR, #UART
	MOVX	A, @DPTR

	; Check value of byte
	CJNE	A, 0x02, receive_byte_no_good
	
	; receive byte was good
	INC	R4
	MOV	A, R4
	ADD	A, #32
	CALL	putchar
	MOV	A, #'.'
	CALL	putchar

	; if we've gotten the whole message, start the crypt mode
	CJNE	R4, #8, start_crypt_mode
	JMP	initiate_loop

no_receive_byte_ready:
	DJNZ	R5, initiate_loop
 	MOV	R4, #0
	MOV	R5, #32
	
	JMP	initiate_loop

receive_byte_no_good:
	; reset the counters
 	MOV	R4, #0
	MOV	R5, #32

	MOV	A, #'q'
	CALL	putchar	

	JMP	initiate_loop

send_response_message:
	; Here we just send a message back, without reading 
	; anything in....
	
	MOV	DPTR, #got_hello_str
	CALL	putstrc

	MOV	R4, #0

respond_loop:
	; wait for interrupt
	JNB	TF0, $
	CLR	TF0
	MOV	TH0, #80h
	MOV	TL0, #00h
		
	MOV	DPTR, #respond_msg
	MOV	A, R4
	MOVC	A, @A+DPTR
	
	MOV	DPTR, #UART
	MOVX	@DPTR, A
	
	INC	R4
	CJNE	R4, #8, respond_loop

	; Go to crypt mode
	JMP	start_crypt_mode

start_crypt_mode:
	MOV	DPTR,#got_respond_str
	CALL	putstrc
	
	RET

hello_msg:
	DB	70h,80h,90h,0a0h,0b0h,0c0h,0d0h,0e0h

respond_msg:
	DB	80h, 70h, 60h, 50h, 40h, 30h, 20h, 10h

receiver_mode_str:
	DB	10,13,'Receive Mode',10,13,0

initiator_mode_str:
	DB	10,13,'Initiate Mode',10,13,0

got_hello_str:	
	DB	10,13,'Got Hello',10,13,0

got_respond_str:
	DB	10,13,'Got Response',10,13,0
