; Copyright (C) 1997, 1998, 1999 Kevin Fu <fubob@mit.edu>
; Copyright (C) 1997, 1998, 1999 Ben Adida <ben@mit.edu>
; Copyright (C) 1997, 1998, 1999 Rodrigo Leroux
;
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License as
; published by the Free Software Foundation; either version 2, or (at
; your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
; USA

$INCLUDE (..\common\REG51FX.INC)
$INCLUDE (VOICE.INC)

;; Declare external functions
EXTRN	CODE	(putchar)

;; Segment definitions
prog	SEGMENT	CODE	; code memory
	USING	0
CSEG AT 8000H
	JMP	START

;------------------------
;Define parameters and RAM locations
N	EQU	9	; n = number of distinct coefficients
NT2	EQU	18	; 2n = number of FIR coefficients
BB	EQU	30H	; base of 18 byte buffer in directly addr. RAM
CBsize	EQU	24000	; size of circular buffer in external memory
CBbase	EQU	9000H	; base of circular buffer in external memory
CBend	EQU	CBbase+CBsize+1	; address of end of circular buffer + 1
DPHend	EQU	HIGH(CBend)
DPLend	EQU	LOW(CBend)

;------------------------
;Multiplier coefficients, computed k multiplied by 1024 to preserve accuracy.
K0	EQU	6	; -6.45 = 1024 k0
K1	EQU	0	; -0.04 = 1024 k1
K2	EQU	21	; 21.17
K3	EQU	20	; 20.13
K4	EQU	30	; -30.01
K5	EQU	72	; -72.06
K6	EQU	0	; -0.04
K7	EQU	199	; 198.74
K8_M1	EQU	125	; 380.55-256
; Possible range: -255*(72+30+6)/512 to (255*(381+199+20+21))/512 = -53 to 308
; Sum of Ks (with correct sign) = 512


;------------------------
; This is the endless loop for measurement and calculation.
; Timer 0 in Mode 2 is used to trigger an A/D reading every 80 microseconds.
; With a 16 MHz clock there are 106 processor cycles available per measurement.
; The following code reads A/D, does FIR calculation, stores every other filtered byte.
; The 1-byte A/D readings are converted to positive values in offset representation.
; DPTR is pointer for storing filtered data in a circular buffer in external memory.
; R0 and P2 are used to access the A/D converter.
; Reading A/D assumed to start next conversion, but other approaches are possible.
; Maximum execution time in cycles is in right column.
; The calculations are organized with ADD and then SUB so there can be at most
; 2 R6 increments and 1 R6 decrement, and thus only 3 macros allow 15 cycles.

	RSEG prog
START:
	MOV	TMOD, #22H	; set timer 0 to auto-reload mode
	MOV	TH0, #150       ; set the period of timer 0 to 80uS
	SETB	TR0             ; start timer 0
	MOV	DPTR,#CBbase	; init circ buffer ptr
MAIN:	;MOV	A,#'R'
;	CALL	putchar
	;MOV	A,#13
	;CALL	putchar

	JNB	TF0,$	; wait for timer to read A/D		2
	CLR	TF0	;					1
	CLR	P2.7	; A15 goes low                          1
	MOVX	A,@R0	; read A/D, cpl ACC.7, save in BB+1	2
	
	CPL	ACC.7	;					1
	MOV	BB+1,A	; save A/D reading			1
	MOV	A,BB+8	; y = x(8)+x(9)				1
	ADD	A,BB+9	;					1
	RRC	A	;divide by 2				1
	MOV	R7,A	;					1
	MOV	R6,#0	;					1
	ADD_KX	8,K8_M1	; y =y + (k8-1)*(x(8)+x(9))		15
	ADD_KX	7,K7	; y =y + k7*(x(7)+x(10))		15
	ADD_KX	3,K3	; y =y + k3*(x(3)+x(14))		14
	ADD_KX	2,K2	; y =y + k2*(x(2)+x(15))		14
;	SUB_KX	6,K6	; k6 = -0.04 so omit			0
	SUB_KX	5,K5	; y =y - k5*(x(5)+x(12))		14
	CLR	P1.7	; flip !convst				1
	SETB    P1.7    ; suppose it is connected to P1.7       1
	SUB_KX	4,K4	; y - k4*(x(4)+x(13))			14
;	SUB_KX	1,K1	; k1 = -0.04 so omit			0
			; maximum cycles so far			101

	JNB	TF0,$	; wait for timer to read A/D		2
	CLR	TF0	;					1
	CLR	P2.7
	MOVX	A,@R0	; read A/D, cpl. sign, put in BB+0	2
	CPL	ACC.7	;					1
	MOV	BB+0,A	;					1
	SUB_KX	0,K0	; y - k0*(x(0)+x(17)); y calculation done	15
	MOV	A,R6	; divide result by 2			1
	CLR	C	;					1
	RRC	A	;					1
	JZ	L2	; jump if no underflow or overflow	2
	JNB	ACC.6,L1	; jump if overflow		2
	CLR	A	; underflow so set A = 0		1
	SJMP	L3	;					2
L1:	MOV	A,#255	; overflow so set A = 255		1	
	SJMP	L3	;	                                2
L2:	MOV	A,R7	; complete division by 2	        1
	RRC	A	;	                                1
L3:	MOVX	@DPTR,A	; store next data byte in circular buffer	2
	INC	DPTR	; set pointer to next position in buffer	2
	MOV	A,#DPLend	; check DPL for end of circular buffer	1
	CJNE	A,DPL,L4	; jump if not at end			2
	MOV	A,#DPHend	; check DPH also			1
	CJNE	A,DPH,L4	;					2
	MOV	DPTR,#CBbase	; reset pointer to base of memory space	2
L4:	MOV	BB+17,BB+15	; shift data in FIR buffer		2
	MOV	BB+16,BB+14	;					2
	MOV	BB+15,BB+13	;					2
	MOV	BB+14,BB+12	;					2
	MOV	BB+13,BB+11	;					2
	MOV	BB+12,BB+10	;					2
	MOV	BB+11,BB+9	;					2
	MOV	BB+10,BB+8	;					2
	MOV	BB+9,BB+7	;					2
	MOV	BB+8,BB+6	;					2
	CLR	P1.7		; flip !convst				1
	SETB    P1.7    	; suppose it is connected to P1.7       1
	MOV	BB+7,BB+5	;					2
	MOV	BB+6,BB+4	;					2
	MOV	BB+5,BB+3	;					2
	MOV	BB+4,BB+2	;					2
	MOV	BB+3,BB+1	;					2
	MOV	BB+2,BB+0	;					2
	JB	P1.1,finished	; jump out of the loop if P1.1 is high  2

			; up to 23 more cycles can be inserted	
	LJMP	MAIN	; repeat forever				2
			; maximum cycles in second half of loop		83

; turn P1.1 (or other pin, if you want) low with a switch to record the voice,
; and turn it high to stop recording. You can do other things, such as play back
; or print out contents in the circular buffer (for debug) in "finished".

finished:
;	MOV	A,#'P'
;	CALL	putchar
;	MOV	A,#13
;	CALL	putchar

	;; Playback time!  
readCB:	JNB	TF0,$	; wait for timer to read A/D		2
	CLR	TF0	;					1

	MOVX	A,@DPTR	; read next circular buffer data byte		2
	CPL	ACC.7	; offset -> 2s complement
	CLR	P2.7	; Enable DAC
	MOVX	@R0, A

	INC	DPTR	; set pointer to next position in buffer	2
	MOV	A,#DPLend	; check DPL for end of circular buffer	1
	CJNE	A,DPL,L5	; jump if not at end			2
	MOV	A,#DPHend	; check DPH also			1
	CJNE	A,DPH,L5	;					2
	MOV	DPTR,#CBbase	; reset pointer to base of memory space	2
L5:	
	JNB	TF0,$	; burn this interrupt cycle
	CLR	TF0
;	NOP
;	NOP
;	NOP
;	NOP
;	CLR	P2.7	; Enable DAC
;	MOVX	@R0, A

	JB	P1.1,finished
	LJMP	main				

END






